import threading
from operator import itemgetter
from itertools import chain

from celery.states import READY_STATES, RECEIVED
from celery.utils.functional import LRUCache
from celery.events.state import Task, TASK_EVENT_TO_STATE

CELERY_DEFAULT_QUEUE = "celery"
CELERY_MISSING_DATA = "undefined"


class CeleryState:
    event_count = 0
    task_count = 0

    def __init__(self, max_tasks_in_memory=10000):
        self.tasks = LRUCache(max_tasks_in_memory)
        self._queue_by_task = {}
        self._mutex = threading.Lock()

    @classmethod
    def _gen_wildcards(self, name):
        chunked = name.split(".")
        res = [name]
        for elem in reversed(chunked):
            chunked.pop()
            res.append(".".join(chunked + ["*"]))
        return res

    @classmethod
    def get_config(self, app):
        res = dict()
        try:
            registered_tasks = app.control.inspect().registered_tasks().values()
            confs = app.control.inspect().conf()
        except Exception:  # pragma: no cover
            return res

        for task_name in set(chain.from_iterable(registered_tasks)):
            for conf in confs.values():
                default = conf.get("task_default_queue", CELERY_DEFAULT_QUEUE)
                if task_name in res and res[task_name] != default:
                    break

                task_wildcard_names = self._gen_wildcards(task_name)
                if "task_routes" in conf:
                    routes = conf["task_routes"]
                    res[task_name] = default
                    for i in task_wildcard_names:
                        if i in routes and "queue" in routes[i]:
                            res[task_name] = routes[i]["queue"]
                            break
                else:
                    res[task_name] = default
        return res

    def _measure_latency(self, evt):
        try:
            prev_evt = self.tasks[evt["uuid"]]
        except KeyError:  # pragma: no cover
            pass
        else:
            if prev_evt.state == RECEIVED:
                return evt["local_received"] - prev_evt.local_received

        return None

    def latency(self, evt):
        group, _, subject = evt["type"].partition("-")
        if group == "task":
            if subject == "started":
                return self._measure_latency(evt)
        return None

    def _event(self, evt, subject):
        tfields = itemgetter("uuid", "hostname", "timestamp", "local_received", "clock")
        get_task = self.tasks.__getitem__
        self.event_count += 1

        (uuid, hostname, timestamp, local_received, clock) = tfields(evt)
        is_client_event = subject == "sent"
        try:
            task, task_created = get_task(uuid), False
        except KeyError:
            task = self.tasks[uuid] = Task(uuid, cluster_state=None)
            task_created = True
        if is_client_event:
            task.client = hostname

        if subject == "received":
            self.task_count += 1

        task.event(subject, timestamp, local_received, evt)

        name = task.name
        if name is not None and "queue" in evt:
            if not name in self._queue_by_task:
                self._queue_by_task[name] = evt["queue"]

        return (task, task_created), subject

    def event(self, evt, subject):
        with self._mutex:
            return self._event(evt, subject)

    def collect(self, evt):
        group, _, subject = evt["type"].partition("-")
        runtime = None
        if group == "task":
            state = TASK_EVENT_TO_STATE[subject]
            if state in READY_STATES:
                try:
                    with self._mutex:
                        name = self.tasks.pop(evt["uuid"]).name or CELERY_MISSING_DATA
                except (KeyError):  # pragma: no cover
                    name = CELERY_MISSING_DATA
                finally:
                    queue = self._queue_by_task.get(name, CELERY_MISSING_DATA)
                    if "runtime" in evt:
                        runtime = evt["runtime"]
                    return (name, state, runtime, queue)
            else:
                self.event(evt, subject)
                try:
                    name = self.tasks[evt["uuid"]].name or CELERY_MISSING_DATA
                except (KeyError):  # pragma: no cover
                    name = CELERY_MISSING_DATA
                finally:
                    queue = self._queue_by_task.get(name, CELERY_MISSING_DATA)
                    return (name, state, runtime, queue)
        return (None, None, None, None)



import collections
from itertools import chain
import logging
import time
import threading

import celery
import celery.states

from .metrics import TASKS, TASKS_RUNTIME, LATENCY, WORKERS
from .state import CeleryState


class TaskThread(threading.Thread):
    """
    MonitorThread is the thread that will collect the data that is later
    exposed from Celery using its eventing system.
    """

    def __init__(self, app, namespace, max_tasks_in_memory, *args, **kwargs):
        self._app = app
        self._namespace = namespace
        self.log = logging.getLogger("task-thread")
        self._state = CeleryState(max_tasks_in_memory=max_tasks_in_memory)
        self._known_states = set()
        self._known_states_names = set()
        self._tasks_started = dict()
        super(TaskThread, self).__init__(*args, **kwargs)

    def run(self):  # pragma: no cover
        self._monitor()

    def _process_event(self, evt):
        latency = self._state.latency(evt)
        if latency is not None:
            LATENCY.labels(namespace=self._namespace).observe(latency)
        (name, state, runtime, queue) = self._state.collect(evt)
        if name is not None:
            if runtime is not None:
                TASKS_RUNTIME.labels(
                    namespace=self._namespace, name=name, queue=queue
                ).observe(runtime)

            TASKS.labels(
                namespace=self._namespace, name=name, state=state, queue=queue
            ).inc()

    def _monitor(self):  # pragma: no cover
        while True:
            try:
                with self._app.connection() as conn:
                    recv = self._app.events.Receiver(
                        conn, handlers={"*": self._process_event}
                    )
                    setup_metrics(self._app, self._namespace)
                    self.log.info("Start capturing events...")
                    recv.capture(limit=None, timeout=None, wakeup=True)
            except Exception:
                self.log.exception("Connection failed")
                setup_metrics(self._app, self._namespace)
                time.sleep(5)


class WorkerMonitoringThread(threading.Thread):
    celery_ping_timeout_seconds = 5
    periodicity_seconds = 5

    def __init__(self, app, namespace, *args, **kwargs):
        self._app = app
        self._namespace = namespace
        self.log = logging.getLogger("workers-thread")
        super(WorkerMonitoringThread, self).__init__(*args, **kwargs)

    def run(self):  # pragma: no cover
        while True:
            self.update_workers_count()
            time.sleep(self.periodicity_seconds)

    def update_workers_count(self):
        try:
            WORKERS.labels(namespace=self._namespace).set(
                len(self._app.control.ping(timeout=self.celery_ping_timeout_seconds))
            )
        except Exception:  # pragma: no cover
            self.log.exception("Error while pinging workers")


class EnableEventsThread(threading.Thread):
    periodicity_seconds = 5

    def __init__(self, app, *args, **kwargs):  # pragma: no cover
        self._app = app
        self.log = logging.getLogger("enable-events-thread")
        super(EnableEventsThread, self).__init__(*args, **kwargs)

    def run(self):  # pragma: no cover
        while True:
            try:
                self.enable_events()
            except Exception:
                self.log.exception("Error while trying to enable events")
            time.sleep(self.periodicity_seconds)

    def enable_events(self):
        self._app.control.enable_events()


def setup_metrics(app, namespace):
    """
    This initializes the available metrics with default values so that
    even before the first event is received, data can be exposed.
    """
    WORKERS.labels(namespace=namespace)
    LATENCY.labels(namespace=namespace)
    config = CeleryState.get_config(app)

    if not config:  # pragma: no cover
        for metric in TASKS.collect():
            for name, labels, cnt in metric.samples:
                TASKS.labels(**labels)
    else:
        for state in celery.states.ALL_STATES:
            for task, queue in config.items():
                TASKS.labels(namespace=namespace, name=task, state=state, queue=queue)



from unittest.mock import patch, MagicMock
from celery_test_utils import BaseTest

import celery
import celery_exporter.monitor
from celery_exporter.core import CeleryExporter

prom_http_server_mock = MagicMock(return_value=None)
setup_metrics_mock = MagicMock(return_value=None)
task_thread_mock = MagicMock(spec=celery_exporter.monitor.TaskThread)
worker_thread_mock = MagicMock(spec=celery_exporter.monitor.WorkerMonitoringThread)
event_thread_mock = MagicMock(spec=celery_exporter.monitor.EnableEventsThread)


@patch("celery.task.control.inspect.registered_tasks", {"worker1": [BaseTest.task]})
@patch(
    "celery_exporter.core.prometheus_client.start_http_server", prom_http_server_mock
)
@patch("celery_exporter.core.setup_metrics", setup_metrics_mock)
@patch("celery_exporter.core.TaskThread", task_thread_mock)
@patch("celery_exporter.core.WorkerMonitoringThread", worker_thread_mock)
@patch("celery_exporter.core.EnableEventsThread", event_thread_mock)
class TestCeleryExporter(BaseTest):
    def setUp(self):
        self.cel_exp = CeleryExporter(
            broker_url="memory://",
            listen_address="127.0.0.1:9090",
            max_tasks=TestCeleryExporter.max_tasks,
            namespace=TestCeleryExporter.namespace,
            enable_events=True,
        )

    def test_setup_metrics(self):
        self.cel_exp.start()
        setup_metrics_mock.assert_called_with(
            self.cel_exp._app, TestCeleryExporter.namespace
        )

    def test_http_server(self):
        self.cel_exp.start()
        prom_http_server_mock.assert_called_with(9090, "127.0.0.1")

    def test_task_thread(self):
        self.cel_exp.start()
        task_thread_mock.assert_called_with(
            self.cel_exp._app,
            TestCeleryExporter.namespace,
            TestCeleryExporter.max_tasks,
        )

    def test_worker_thread(self):
        self.cel_exp.start()
        worker_thread_mock.assert_called_with(
            self.cel_exp._app, TestCeleryExporter.namespace
        )

    def test_event_thread(self):
        self.cel_exp.start()
        event_thread_mock.assert_called_with(self.cel_exp._app)



import json
import logging
import os
import signal
import sys
import time

import click
from .core import CeleryExporter

__VERSION__ = (2, 0, 0)

LOG_FORMAT = "[%(asctime)s] %(name)s:%(levelname)s: %(message)s"


@click.command(context_settings={"auto_envvar_prefix": "CELERY_EXPORTER"})
@click.option(
    "--broker-url",
    "-b",
    type=str,
    show_default=True,
    show_envvar=True,
    default="redis://redis:6379/0",
    help="URL to the Celery broker.",
)
@click.option(
    "--listen-address",
    "-l",
    type=str,
    show_default=True,
    show_envvar=True,
    default="0.0.0.0:9540",
    help="Address the HTTPD should listen on.",
)
@click.option(
    "--max-tasks",
    "-m",
    type=int,
    show_default=True,
    show_envvar=True,
    default="10000",
    help="Tasks cache size.",
)
@click.option(
    "--namespace",
    "-n",
    type=str,
    show_default=True,
    show_envvar=True,
    default="celery",
    help="Namespace for metrics.",
)
@click.option(
    "--transport-options",
    type=str,
    allow_from_autoenv=False,
    help="JSON object with additional options passed to the underlying transport.",
)
@click.option(
    "--enable-events",
    is_flag=True,
    allow_from_autoenv=False,
    help="Periodically enable Celery events.",
)
@click.option(
    "--tz", type=str, allow_from_autoenv=False, help="Timezone used by the celery app."
)
@click.option(
    "--verbose", is_flag=True, allow_from_autoenv=False, help="Enable verbose logging."
)
@click.version_option(version=".".join([str(x) for x in __VERSION__]))
def main(
    broker_url,
    listen_address,
    max_tasks,
    namespace,
    transport_options,
    enable_events,
    tz,
    verbose,
):  # pragma: no cover

    if verbose:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    if tz:
        os.environ["TZ"] = tz
        time.tzset()

    if transport_options:
        try:
            transport_options = json.loads(transport_options)
        except ValueError:
            print(
                "Error parsing broker transport options from JSON '{}'".format(
                    transport_options
                ),
                file=sys.stderr,
            )
            sys.exit(1)

    celery_exporter = CeleryExporter(
        broker_url,
        listen_address,
        max_tasks,
        namespace,
        transport_options,
        enable_events,
    )
    celery_exporter.start()

    def shutdown(signum, frame):  # pragma: no cover
        """
        Shutdown is called if the process receives a TERM/INT signal.
        """
        logging.info("Shutting down")
        sys.exit(0)

    signal.signal(signal.SIGINT, shutdown)
    signal.signal(signal.SIGTERM, shutdown)

    while True:
        signal.pause()


if __name__ == "__main__":  # pragma: no cover
    main()  # pylint: disable=E1120



from time import time

import celery
import celery.states

from celery.events import Event
from celery.utils import uuid
from prometheus_client import REGISTRY
from unittest import TestCase
from unittest.mock import patch

from celery_exporter.monitor import (
    WorkerMonitoringThread,
    TaskThread,
    EnableEventsThread,
    setup_metrics,
)

from celery_exporter.state import CELERY_MISSING_DATA, CeleryState

from celery_test_utils import BaseTest, get_celery_app


class TestMockedCelery(BaseTest):
    def setUp(self):
        self.app = get_celery_app()
        with patch("celery.task.control.inspect.conf") as tasks:
            with patch("celery.task.control.inspect.registered_tasks") as registered:
                tasks.return_value = {
                    "celery@d6f95e9e24fc": {
                        "task_routes": {"my_task": {}, "trial": {"queue": "deadbeef"}}
                    },
                    "celery@adsqas78e891": {
                        "task_routes": {"my_task": {}, "trial": {"queue": "deadbeef"}}
                    },
                    "celery@12311847jsa2": {},
                }
                registered.return_value = {"celery@d6f95e9e24fc": [self.task, "trial"]}
                setup_metrics(self.app, self.namespace)  # reset metrics

    def test_initial_metric_values(self):
        self._assert_task_states(celery.states.ALL_STATES, 0)
        assert (
            REGISTRY.get_sample_value(
                "celery_workers", labels=dict(namespace=self.namespace)
            )
            == 0
        )
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_count",
                labels=dict(namespace=self.namespace),
            )
            == 0
        )
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_sum",
                labels=dict(namespace=self.namespace),
            )
            == 0
        )

    def test_workers_count(self):
        assert (
            REGISTRY.get_sample_value(
                "celery_workers", labels=dict(namespace=self.namespace)
            )
            == 0
        )

        with patch.object(self.app.control, "ping") as mock_ping:
            w = WorkerMonitoringThread(app=self.app, namespace=self.namespace)

            mock_ping.return_value = []
            w.update_workers_count()
            assert (
                REGISTRY.get_sample_value(
                    "celery_workers", labels=dict(namespace=self.namespace)
                )
                == 0
            )

            mock_ping.return_value = [0]  # 1 worker
            w.update_workers_count()
            assert (
                REGISTRY.get_sample_value(
                    "celery_workers", labels=dict(namespace=self.namespace)
                )
                == 1
            )

            mock_ping.return_value = [0, 0]  # 2 workers
            w.update_workers_count()
            assert (
                REGISTRY.get_sample_value(
                    "celery_workers", labels=dict(namespace=self.namespace)
                )
                == 2
            )

            mock_ping.return_value = []
            w.update_workers_count()
            assert (
                REGISTRY.get_sample_value(
                    "celery_workers", labels=dict(namespace=self.namespace)
                )
                == 0
            )

    def test_tasks_events(self):
        task_uuid = uuid()
        hostname = "myhost"
        local_received = time()
        latency_before_started = 123.45
        runtime = 234.5

        m = TaskThread(
            app=self.app, namespace=self.namespace, max_tasks_in_memory=self.max_tasks
        )

        self._assert_task_states(celery.states.ALL_STATES, 0)
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_count",
                labels=dict(namespace=self.namespace),
            )
            == 0
        )
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_sum",
                labels=dict(namespace=self.namespace),
            )
            == 0
        )

        m._process_event(
            Event(
                "task-sent",
                uuid=task_uuid,
                name=self.task,
                queue=self.queue,
                args="()",
                kwargs="{}",
                retries=0,
                eta=None,
                hostname=hostname,
                clock=0,
                local_received=local_received,
            )
        )
        self._assert_all_states({celery.states.PENDING})

        m._process_event(
            Event(
                "task-received",
                uuid=task_uuid,
                name=self.task,
                args="()",
                kwargs="{}",
                retries=0,
                eta=None,
                hostname=hostname,
                clock=0,
                local_received=local_received,
            )
        )
        self._assert_all_states({celery.states.PENDING, celery.states.RECEIVED})

        m._process_event(
            Event(
                "task-started",
                uuid=task_uuid,
                hostname=hostname,
                clock=1,
                name=self.task,
                local_received=local_received + latency_before_started,
            )
        )
        self._assert_all_states(
            {celery.states.PENDING, celery.states.RECEIVED, celery.states.STARTED}
        )

        m._process_event(
            Event(
                "task-succeeded",
                uuid=task_uuid,
                result="42",
                runtime=runtime,
                hostname=hostname,
                clock=2,
                local_received=local_received + latency_before_started + runtime,
            )
        )
        self._assert_all_states(
            {
                celery.states.PENDING,
                celery.states.RECEIVED,
                celery.states.STARTED,
                celery.states.SUCCESS,
            }
        )

        m._process_event(
            Event(
                "task-started",
                uuid=task_uuid,
                result="42",
                runtime=runtime,
                hostname=hostname,
                clock=2,
                local_received=local_received + latency_before_started + runtime,
            )
        )
        self._assert_task_states({celery.states.STARTED}, 1)

        m._process_event(
            Event(
                "notatask-sent",
                uuid=task_uuid,
                name=self.task,
                args="()",
                kwargs="{}",
                retries=0,
                eta=None,
                hostname=hostname,
                clock=0,
                local_received=local_received,
            )
        )
        self._assert_task_states({celery.states.PENDING}, 1)

        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_count",
                labels=dict(namespace=self.namespace),
            )
            == 1
        )
        self.assertAlmostEqual(
            REGISTRY.get_sample_value(
                "celery_tasks_latency_seconds_sum",
                labels=dict(namespace=self.namespace),
            ),
            latency_before_started,
        )
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_runtime_seconds_count",
                labels=dict(namespace=self.namespace, name=self.task, queue=self.queue),
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_runtime_seconds_sum",
                labels=dict(namespace=self.namespace, name=self.task, queue=self.queue),
            )
            == 234.5
        )

        m._process_event(
            Event(
                "task-succeeded",
                uuid=uuid(),
                result="42",
                runtime=runtime,
                hostname=hostname,
                clock=2,
                local_received=local_received + latency_before_started + runtime,
            )
        )

        assert (
            REGISTRY.get_sample_value(
                "celery_tasks_total",
                labels={
                    "namespace": self.namespace,
                    "state": celery.states.STARTED,
                    "name": CELERY_MISSING_DATA,
                    "queue": CELERY_MISSING_DATA,
                },
            )
            == 1
        )

    def test_enable_events(self):
        with patch.object(self.app.control, "enable_events") as mock_enable_events:
            e = EnableEventsThread(app=self.app)
            e.enable_events()
            mock_enable_events.assert_called_once_with()

    def test_gen_wildcards(self):
        strings = {
            "aaa.bbb.ccc.ddd": [
                "aaa.bbb.ccc.ddd",
                "aaa.bbb.ccc.*",
                "aaa.bbb.*",
                "aaa.*",
                "*",
            ],
            "aaa.bbb": ["aaa.bbb", "aaa.*", "*"],
            "aaa": ["aaa", "*"],
        }
        for case, expectation in strings.items():
            result = CeleryState._gen_wildcards(case)
            assert result == expectation

    def _assert_task_states(self, states, cnt):
        for state in states:
            task_by_name_label = dict(
                namespace=self.namespace, name=self.task, state=state, queue=self.queue
            )

            assert (
                REGISTRY.get_sample_value(
                    "celery_tasks_total", labels=task_by_name_label
                )
                == cnt
            )

    def _assert_all_states(self, exclude):
        self._assert_task_states(celery.states.ALL_STATES - exclude, 0)
        self._assert_task_states(exclude, 1)



import celery
import logging
import prometheus_client
from .monitor import (
    TaskThread,
    WorkerMonitoringThread,
    EnableEventsThread,
    setup_metrics,
)

__all__ = ("CeleryExporter",)


class CeleryExporter:
    def __init__(
        self,
        broker_url,
        listen_address,
        max_tasks=10000,
        namespace="celery",
        transport_options=None,
        enable_events=False,
    ):
        self._listen_address = listen_address
        self._max_tasks = max_tasks
        self._namespace = namespace
        self._enable_events = enable_events

        self._app = celery.Celery(broker=broker_url)
        self._app.conf.broker_transport_options = transport_options or {}

    def start(self):

        setup_metrics(self._app, self._namespace)

        self._start_httpd()

        t = TaskThread(
            app=self._app,
            namespace=self._namespace,
            max_tasks_in_memory=self._max_tasks,
        )
        t.daemon = True
        t.start()

        w = WorkerMonitoringThread(app=self._app, namespace=self._namespace)
        w.daemon = True
        w.start()

        if self._enable_events:
            e = EnableEventsThread(app=self._app)
            e.daemon = True
            e.start()

    def _start_httpd(self):  # pragma: no cover
        """
        Starts the exposing HTTPD using the addr provided in a separate
        thread.
        """
        host, port = self._listen_address.split(":")
        logging.info("Starting HTTPD on {}:{}".format(host, port))
        prometheus_client.start_http_server(int(port), host)



import io

from setuptools import setup

long_description = "See https://github.com/OvalMoney/celery-exporter"
with io.open("README.md", encoding="utf-8") as fp:
    long_description = fp.read()

setup(
    name="celery-exporter",
    description="Prometheus metrics exporter for Celery",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="1.3.3",
    author="Fabio Todaro",
    license="MIT",
    author_email="fbregist@gmail.com",
    url="https://github.com/OvalMoney/celery-exporter",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    packages=["celery_exporter"],
    install_requires=["click>=7" "celery>=4", "prometheus_client>=0.0.20"],
    entry_points={
        "console_scripts": ["celery-exporter = celery_exporter.__main__:main"]
    },
)



import prometheus_client

TASKS = prometheus_client.Counter(
    "celery_tasks_total",
    "Number of task events.",
    ["namespace", "name", "state", "queue"],
)
TASKS_RUNTIME = prometheus_client.Histogram(
    "celery_tasks_runtime_seconds", "Task runtime.", ["namespace", "name", "queue"]
)
LATENCY = prometheus_client.Histogram(
    "celery_tasks_latency_seconds",
    "Time between a task is received and started.",
    ["namespace"],
)
WORKERS = prometheus_client.Gauge(
    "celery_workers", "Number of alive workers", ["namespace"]
)
